#include <msp/core/maputils.h>
#include <msp/gltk/button.h>
#include <msp/gltk/style.h>
#include "engineer.h"
#include "libr2c2/train.h"
#include "newtraindialog.h"
#include "traindialog.h"
#include "trainlistdialog.h"

using namespace std;
using namespace Msp;
using namespace R2C2;

class TrainItem: public GLtk::List::MultiColumnItem
{
public:
	typedef R2C2::Train *ValueType;

	TrainItem(ValueType);
};


class DetectedLocoItem: public GLtk::List::MultiColumnItem
{
public:
	typedef const Driver::DetectedLocomotive *ValueType;

	DetectedLocoItem(ValueType);
};


TrainListDialog::TrainListDialog(Engineer &e):
	DynamicDialog(e.get_user_interface()),
	engineer(e),
	layout(engineer.get_layout())
{
	Loader::WidgetMap widgets;
	DataFile::load(*this, "data/trainlistdialog.ui", widgets);

	lst_trains = dynamic_cast<GLtk::List *>(get_item(widgets, "lst_trains"));
	lst_trains->set_data(trains);
	lst_trains->set_item_type<TrainItem>();

	lst_imported = dynamic_cast<GLtk::List *>(get_item(widgets, "lst_imported"));
	lst_imported->set_data(detected_locos);
	lst_imported->set_item_type<DetectedLocoItem>();

	dynamic_cast<GLtk::Toggle *>(get_item(widgets, "tgl_trains"))->signal_toggled.connect(sigc::bind(sigc::mem_fun(this, &TrainListDialog::tab_toggled), lst_trains));
	dynamic_cast<GLtk::Toggle *>(get_item(widgets, "tgl_import"))->signal_toggled.connect(sigc::bind(sigc::mem_fun(this, &TrainListDialog::tab_toggled), lst_imported));

	dynamic_cast<GLtk::Button *>(get_item(widgets, "btn_new"))->signal_clicked.connect(sigc::mem_fun(this, &TrainListDialog::new_clicked));
	dynamic_cast<GLtk::Button *>(get_item(widgets, "btn_show"))->signal_clicked.connect(sigc::mem_fun(this, &TrainListDialog::show_clicked));

	const map<unsigned, Train *> &ltrains = layout.get_trains();
	for(map<unsigned, Train *>::const_iterator i=ltrains.begin(); i!=ltrains.end(); ++i)
	{
		trains.append(i->second);
		i->second->signal_name_changed.connect(sigc::hide(sigc::bind(sigc::mem_fun(this, &TrainListDialog::train_name_changed), sigc::ref(*i->second))));
	}

	if(layout.has_driver())
	{
		Driver &driver = layout.get_driver();
		driver.signal_locomotive_detected.connect(sigc::mem_fun(this, &TrainListDialog::locomotive_detected));
		driver.signal_locomotive_gone.connect(sigc::mem_fun(this, &TrainListDialog::locomotive_gone));

		unsigned i = 0;
		while(const Driver::DetectedLocomotive *loco = driver.enumerate_detected_locos(i++))
			detected_locos.append(loco);
	}
	layout.signal_train_added.connect(sigc::mem_fun(this, &TrainListDialog::train_added));
	layout.signal_train_removed.connect(sigc::mem_fun(this, &TrainListDialog::train_removed));
}

void TrainListDialog::new_clicked()
{
	NewTrainDialog *dlg = new NewTrainDialog(engineer);
	if(lst_imported->is_visible())
	{
		int selected = lst_imported->get_selected_index();
		if(selected>=0)
			dlg->prefill(*detected_locos.get(selected));
	}
	find_ancestor<GLtk::Root>()->add(*dlg);
	dlg->autosize();
}

void TrainListDialog::show_clicked()
{
	int index = lst_trains->get_selected_index();
	if(index>=0)
	{
		TrainDialog *dlg = new TrainDialog(engineer, *trains.get(index));
		find_ancestor<GLtk::Root>()->add(*dlg);
	}
}

void TrainListDialog::tab_toggled(bool value, GLtk::Widget *wdg)
{
	wdg->set_visible(value);
}

void TrainListDialog::locomotive_detected(const Driver::DetectedLocomotive &loco)
{
	detected_locos.append(&loco);
}

void TrainListDialog::locomotive_gone(const Driver::DetectedLocomotive &loco)
{
	int index = detected_locos.find(&loco);
	if(index>=0)
		detected_locos.remove(index);
}

void TrainListDialog::train_added(Train &train)
{
	unsigned n_items = trains.size();
	unsigned i;
	for(i=0; i<n_items; ++i)
		if(trains.get(i)->get_address()>train.get_address())
			break;
	trains.insert(i, &train);

	train.signal_name_changed.connect(sigc::hide(sigc::bind(sigc::mem_fun(this, &TrainListDialog::train_name_changed), sigc::ref(train))));
}

void TrainListDialog::train_removed(Train &train)
{
	int i = trains.find(&train);
	if(i>=0)
		trains.remove(i);
}

void TrainListDialog::train_name_changed(Train &train)
{
	trains.refresh(&train);
}

bool TrainListDialog::save_state(DataFile::Statement &st) const
{
	st.keyword = "trainlistdialog";
	save_position(st.sub);

	return true;
}


TrainItem::TrainItem(ValueType train)
{
	add(*new GLtk::Label(train->get_protocol()));
	add(*new GLtk::Label(lexical_cast<string>(train->get_address())));
	add(*new GLtk::Label(train->get_name()));
}


DetectedLocoItem::DetectedLocoItem(ValueType loco)
{
	add(*new GLtk::Label(loco->protocol));
	add(*new GLtk::Label(lexical_cast<string>(loco->address)));
	add(*new GLtk::Label(loco->name));
}
