#include <algorithm>
#include <msp/strings/format.h>
#include "aicontrol.h"
#include "clock.h"
#include "layout.h"
#include "timetable.h"
#include "train.h"
#include "trainrouter.h"
#include "zone.h"

using namespace std;
using namespace Msp;

namespace R2C2 {

Timetable::Timetable(Train &t):
	TrainAI(t),
	current_row(rows.end()),
	update_pending(false),
	sync_to_clock(true),
	late_arrival(false),
	next_depart(rows.end())
{
	if(!train.get_ai_of_type<AIControl>())
		new AIControl(train);
	if(!train.get_ai_of_type<TrainRouter>())
		new TrainRouter(train);

	train.signal_ai_event.connect(sigc::mem_fun(this, &Timetable::event));
	train.get_layout().get_clock().signal_discontinuity.connect(sigc::mem_fun(this, &Timetable::clock_discontinuity));
}

void Timetable::append_row(const Row &r)
{
	insert_row(rows.size(), r);
}

void Timetable::insert_row(unsigned i, const Row &r)
{
	if(i>rows.size())
		throw out_of_range("Timetable::insert_row");

	list<Row>::iterator j = rows.begin();
	advance(j, i);
	j = rows.insert(j, r);
	signal_row_added.emit(i, *j);

	check_update(j);
}

void Timetable::modify_row(unsigned i, const Row &r)
{
	if(i>=rows.size())
		throw out_of_range("Timetable::remove_row");

	list<Row>::iterator j = rows.begin();
	advance(j, i);
	*j = r;
	signal_row_modified.emit(i, r);

	check_update(j);
}

void Timetable::remove_row(unsigned i)
{
	if(i>=rows.size())
		throw out_of_range("Timetable::remove_row");

	list<Row>::iterator j = rows.begin();
	advance(j, i);

	check_update(j);
	if(j==current_row)
		--current_row;

	rows.erase(j);
	signal_row_removed.emit(i);
}

const Timetable::Row &Timetable::get_row(unsigned i) const
{
	if(i>=rows.size())
		throw out_of_range("Timetable::get_row");

	list<Row>::const_iterator j = rows.begin();
	advance(j, i);
	return *j;
}

void Timetable::tick(const Time::TimeDelta &dt)
{
	if(update_pending && !train.get_block_allocator().is_active())
		update_route();

	if(next_depart!=rows.end() && next_depart!=current_row && passed_row(*next_depart, dt))
		late_arrival = true;
}

bool Timetable::passed_row(const Row &row, const Time::TimeDelta &dt) const
{
	const Clock &clock = train.get_layout().get_clock();

	Time::TimeDelta t = clock.get_current_time();
	if(t<row.time)
		t += Time::day;

	Time::TimeDelta b = t-dt*clock.get_rate();
	return b<row.time;
}

void Timetable::save(list<DataFile::Statement> &st) const
{
	for(list<Row>::const_iterator i=rows.begin(); i!=rows.end(); ++i)
	{
		DataFile::Statement ss("row");
		i->save(ss.sub);
		st.push_back(ss);
	}
}

void Timetable::check_update(const list<Row>::const_iterator &i)
{
	for(list<Row>::const_iterator j=current_row; (j!=rows.end() && j!=i); ++j)
		if(j->type==ARRIVE)
			return;
	update_pending = true;
}

list<Timetable::Row>::iterator Timetable::find_trip(const list<Row>::iterator &begin, list<Row>::iterator *arrive)
{
	list<Row>::iterator i = find_if(begin, rows.end(), RowTypeMatch(DEPART));
	if(i==rows.end())
		return i;

	list<Row>::iterator j = find_if(i, rows.end(), RowTypeMatch(ARRIVE));
	if(j==rows.end())
		return j;

	if(arrive)
		*arrive = j;
	return i;
}

void Timetable::update_route()
{
	update_pending = false;
	if(rows.empty())
		return;

	const Clock &clock = train.get_layout().get_clock();

	if(sync_to_clock)
	{
		sync_to_clock = false;
		current_row = rows.begin();
		for(list<Row>::iterator i=rows.begin(); i!=rows.end(); ++i)
			if(i->type==DEPART && i->time>=clock.get_current_time())
			{
				current_row = i;
				break;
			}
	}

	list<Row>::iterator arrive;
	list<Row>::iterator depart = find_trip(current_row, &arrive);
	if(depart==rows.end())
	{
		depart = find_trip(rows.begin(), &arrive);
		if(depart==rows.end())
		{
			current_row = rows.end();
			return;
		}
	}

	train.ai_message(Message("clear-route"));

	current_row = depart;
	for(list<Row>::const_iterator i=depart; i!=rows.end(); ++i)
	{
		if(i->type==DEPART)
		{
			Time::TimeDelta dt;
			if(late_arrival)
				dt = Time::min;
			else
			{
				dt = i->time-clock.get_current_time();
				while(dt<Time::zero)
					dt += Time::day;
			}
			train.ai_message(Message("set-departure-delay", dt/clock.get_rate()));
		}
		else
		{
			train.ai_message(Message("add-waypoint", TrainRouter::Waypoint(*i->target, i->direction)));
			if(i->type==ARRIVE)
				break;
		}
	}

	next_depart = find_trip(arrive, 0);
	if(next_depart==rows.end())
		next_depart = find_trip(rows.begin(), 0);
	if(next_depart!=rows.end())
	{
		Time::TimeDelta dt = next_depart->time-depart->time;
		while(dt<=Time::zero)
			dt += Time::day;
		train.ai_message(Message("set-trip-duration", dt/clock.get_rate()));
	}

	late_arrival = false;
}

void Timetable::event(TrainAI &, const Message &msg)
{
	if(msg.type=="departed")
	{
		if(current_row->type==DEPART)
			++current_row;
	}
	else if(msg.type=="arrived")
	{
		if(current_row->type==ARRIVE)
			record_time();
		update_pending = true;
	}
	else if(msg.type=="waypoint-reached")
	{
		const TrackChain *wp = msg.value.value<const TrackChain *>();
		if(current_row->type==THROUGH && current_row->target==wp)
		{
			record_time();
			++current_row;
		}
	}
}

void Timetable::record_time()
{
	current_row->time = train.get_layout().get_clock().get_current_time();
	unsigned i = distance(rows.begin(), current_row);
	signal_row_modified.emit(i, *current_row);
}

void Timetable::clock_discontinuity()
{
	update_pending = true;
	sync_to_clock = true;
}


Timetable::Row::Row():
	type(ARRIVE),
	target(0),
	direction(TrackChain::UNSPECIFIED)
{ }

void Timetable::Row::save(list<DataFile::Statement> &st) const
{
	st.push_back((DataFile::Statement("type"), type));
	st.push_back((DataFile::Statement("time"), time.raw()));
	st.push_back(target->save_reference());
	if(direction)
		st.push_back((DataFile::Statement("direction"), direction));
}


Timetable::Loader::Loader(Timetable &t, Layout &l):
	DataFile::ObjectLoader<Timetable>(t),
	layout(l)
{
	add("row", &Loader::row);
}

void Timetable::Loader::row()
{
	Row r;
	load_sub(r, layout);
	obj.rows.push_back(r);
	obj.update_pending = true;
}


Timetable::Row::Loader::Loader(Row &r, Layout &l):
	DataFile::ObjectLoader<Timetable::Row>(r),
	layout(l)
{
	add("block", &Loader::block);
	add("direction", &Row::direction);
	add("time", &Loader::time);
	add("type", &Row::type);
	add("zone", &Loader::zone);
	add("zone", &Loader::zone_numbered);
}

void Timetable::Row::Loader::block(unsigned id)
{
	obj.target = &layout.get_block(id);
}

void Timetable::Row::Loader::time(Time::RawTime t)
{
	obj.time = Time::TimeDelta(t);
}

void Timetable::Row::Loader::zone(const string &name)
{
	zone_numbered(name, 0);
}

void Timetable::Row::Loader::zone_numbered(const string &name, unsigned number)
{
	obj.target = &layout.get_zone(name, number);
}


void operator<<(LexicalConverter &conv, Timetable::RowType rt)
{
	switch(rt)
	{
	case Timetable::ARRIVE: conv.result("ARRIVE"); return;
	case Timetable::DEPART: conv.result("DEPART"); return;
	case Timetable::THROUGH: conv.result("THROUGH"); return;
	default: throw lexical_error(format("conversion of RowType(%d) to string", rt));
	}
}

void operator>>(const LexicalConverter &conv, Timetable::RowType &rt)
{
	if(conv.get()=="ARRIVE")
		rt = Timetable::ARRIVE;
	else if(conv.get()=="DEPART")
		rt = Timetable::DEPART;
	else if(conv.get()=="THROUGH")
		rt = Timetable::THROUGH;
	else
		throw lexical_error(format("conversion of '%s' to RowType", conv.get()));
}

} // namespace R2C2
