#include <cmath>
#include <msp/core/maputils.h>
#include <msp/strings/format.h>
#include "block.h"
#include "catalogue.h"
#include "driver.h"
#include "layout.h"
#include "track.h"
#include "trackattachment.h"
#include "tracktype.h"

using namespace std;
using namespace Msp;

namespace {

struct AttachmentCompare
{
	unsigned entry;

	AttachmentCompare(unsigned e): entry(e) { }

	bool operator()(const R2C2::TrackAttachment *a1, const R2C2::TrackAttachment *a2) const
	{ return a1->get_offset_from_endpoint(entry)<a2->get_offset_from_endpoint(entry); }
};

}

namespace R2C2 {

Track::Track(Layout &l, const TrackType &t):
	Object(l),
	type(t),
	block(0),
	slope(0),
	flex(false),
	turnout_addr(0),
	sensor_addr(0),
	links(type.get_endpoints().size()),
	active_path(0),
	path_changing(false),
	preferred_exit(-1)
{
	if(type.is_turnout())
	{
		turnout_addr = layout.allocate_turnout_address();

		if(layout.has_driver())
		{
			Driver &driver = layout.get_driver();
			turnout_id = driver.add_turnout(turnout_addr, type);
			driver.signal_turnout.connect(sigc::mem_fun(this, &Track::turnout_event));
			driver.signal_turnout_failed.connect(sigc::mem_fun(this, &Track::turnout_failed));
		}
	}

	layout.add(*this);

	for(unsigned paths = type.get_paths(); !(paths&1); ++active_path, paths>>=1) ;
}

Track::~Track()
{
	break_links();
	if(layout.has_driver() && turnout_id)
		layout.get_driver().remove_turnout(turnout_id);
	layout.remove(*this);
}

Track *Track::clone(Layout *to_layout) const
{
	Track *track = new Track((to_layout ? *to_layout : layout), type);
	track->set_position(position);
	track->set_rotation(rotation);
	return track;
}

void Track::set_block(Block *b)
{
	if(b && !b->has_track(*this))
		throw logic_error("track not in block");
	if(!b && block && block->has_track(*this))
		throw logic_error("track still in block");

	block = b;
}

Block &Track::get_block() const
{
	if(!block)
		throw logic_error("!block");

	return *block;
}

void Track::set_position(const Vector &p)
{
	position = p;
	signal_moved.emit();
	propagate_slope();
}

void Track::set_rotation(const Angle &r)
{
	rotation = wrap_positive(r);
	signal_moved.emit();
}

void Track::set_tilt(const Angle &t)
{
	if(links.size()!=2)
		return;

	tilt = t;
	slope = tan(tilt)*type.get_path_length(0);
	signal_moved.emit();
	propagate_slope();
}

void Track::set_flex(bool f)
{
	flex = f;
}

void Track::propagate_slope()
{
	for(vector<Track *>::const_iterator i=links.begin(); i!=links.end(); ++i)
		if(*i)
			(*i)->check_slope();
}

void Track::check_slope()
{
	if(links.size()!=2)
		return;

	if(links[0] && links[1])
	{
		Vector epp0 = links[0]->get_snap_node(links[0]->get_link_slot(*this)).position;
		Vector epp1 = links[1]->get_snap_node(links[1]->get_link_slot(*this)).position;
		position.z = epp0.z;
		slope = epp1.z-position.z;
		tilt = Geometry::atan(slope/type.get_path_length(0));
	}
	else
	{
		if(links[0])
		{
			Vector epp = links[0]->get_snap_node(links[0]->get_link_slot(*this)).position;
			position.z = epp.z;
		}
		else if(links[1])
		{
			Vector epp = links[1]->get_snap_node(links[1]->get_link_slot(*this)).position;
			position.z = epp.z-slope;
		}
	}

	signal_moved.emit();
}

void Track::set_turnout_address(unsigned a)
{
	if(!type.is_turnout())
		throw logic_error("not a turnout");
	if(!a)
		throw invalid_argument("Track::set_turnout_id");

	Driver *driver = (layout.has_driver() ? &layout.get_driver() : 0);

	if(driver && turnout_id)
		driver->remove_turnout(turnout_id);
	turnout_addr = a;
	layout.create_blocks(*this);
	layout.update_routes();
	if(driver && turnout_addr)
		turnout_id = driver->add_turnout(turnout_addr, type);
	else
		turnout_id = 0;
}

void Track::set_sensor_address(unsigned a)
{
	if(type.is_turnout())
		throw logic_error("is a turnout");

	sensor_addr = a;
	layout.create_blocks(*this);
}

void Track::set_preferred_exit(int e)
{
	preferred_exit = e;
}

void Track::set_active_path(unsigned p)
{
	if(!type.is_turnout())
		throw logic_error("not a turnout");
	if(!(type.get_paths()&(1<<p)))
		throw invalid_argument("Track::set_active_path");

	if(active_path==p)
		return;

	signal_path_changing.emit(p);
	path_changing = true;
	layout.get_driver().set_turnout(turnout_id, p);
}

float Track::get_path_length(int p) const
{
	if(p<0)
		p = active_path;
	return type.get_path_length(p);
}

OrientedPoint Track::get_point(unsigned epi, unsigned path, float d) const
{
	OrientedPoint p = type.get_point(epi, path, d);

	p.position = position+rotated_vector(p.position, rotation);
	p.rotation += rotation;
	if(type.get_endpoints().size()==2)
	{
		float dz = tan(tilt)*d;
		if(epi==0)
		{
			p.position.z += dz;
			p.tilt = tilt;
		}
		else
		{
			p.position.z += slope-dz;
			p.tilt = -tilt;
		}
	}

	return p;
}

OrientedPoint Track::get_point(unsigned epi, float d) const
{
	return get_point(epi, active_path, d);
}

unsigned Track::get_n_snap_nodes() const
{
	return type.get_endpoints().size();
}

Snap Track::get_snap_node(unsigned i) const
{
	const vector<TrackType::Endpoint> &eps = type.get_endpoints();
	if(i>=eps.size())
		throw out_of_range("Track::get_snap_node");

	Snap result;
	const TrackType::Endpoint &ep = eps[i];

	result.position = position+rotated_vector(ep.pos, rotation);
	if(eps.size()==2 && i==1)
		result.position.z += slope;

	result.rotation = rotation+ep.dir;

	return result;
}

bool Track::snap(Snap &sn, float limit, SnapType what) const
{
	if(Object::snap(sn, limit, what))
		return true;

	if(what&SNAP_SEGMENT)
	{
		Vector local = rotated_vector(sn.position-position, -rotation);

		OrientedPoint np = type.get_nearest_point(local);
		Vector span = local-np.position;
		if(dot(span, span)<=limit*limit)
		{
			sn.position = position+rotated_vector(np.position, rotation);
			sn.rotation = np.rotation+rotation;
			return true;
		}
	}

	return false;
}

SnapType Track::get_default_snap_type_to(const Object &other) const
{
	if(dynamic_cast<const Track *>(&other))
		return SNAP_NODE;

	return NO_SNAP;
}

unsigned Track::get_n_link_slots() const
{
	return links.size();
}

Track *Track::get_link(unsigned i) const
{
	if(i>=links.size())
		throw out_of_range("Track::get_link");

	return links[i];
}

int Track::get_link_slot(const Object &other) const
{
	for(unsigned i=0; i<links.size(); ++i)
		if(links[i]==&other)
			return i;

	return -1;
}

bool Track::link_to(Object &other)
{
	Track *otrack = dynamic_cast<Track *>(&other);
	if(!otrack)
		return false;

	float gauge_ratio = otrack->get_type().get_gauge()/type.get_gauge();
	if(gauge_ratio<0.99 || gauge_ratio>1.01)
		return false;

	float limit = type.get_gauge();
	if(!flex && !otrack->get_flex())
		limit /= 10;
	limit *= limit;

	unsigned nsn = get_n_snap_nodes();
	unsigned other_nsn = other.get_n_snap_nodes();
	for(unsigned i=0; i<nsn; ++i)
	{
		Snap sn = get_snap_node(i);
		for(unsigned j=0; j<other_nsn; ++j)
		{
			Snap osn = other.get_snap_node(j);
			Vector span = osn.position-sn.position;
			Angle da = wrap_balanced(osn.rotation-sn.rotation-Angle::half_turn());

			if(dot(span, span)<limit && abs(da).radians()<0.01)
			{
				break_link(i);
				otrack->break_link(j);
				links[i] = otrack;
				otrack->links[j] = this;
				check_slope();
				layout.create_blocks(*this);

				signal_link_changed.emit(i, otrack);
				otrack->signal_link_changed.emit(j, this);
				return true;
			}
		}
	}

	return false;
}

bool Track::break_link(unsigned i)
{
	if(i>=links.size())
		throw out_of_range("Track::break_link");

	Track *other = links[i];
	if(!other)
		return false;

	links[i] = 0;
	if(!other->break_link(*this))
	{
		/* If the call doesn't succeed, it means that the other track already
		broke the link and is calling us right now.  Recreate blocks in the inner
		call so it occurs before any signals are emitted. */
		layout.create_blocks(*this);
	}

	signal_link_changed.emit(i, 0);

	return true;
}

void Track::add_attachment(TrackAttachment &a)
{
	if(find(attachments.begin(), attachments.end(), &a)!=attachments.end())
		throw key_error(&a);
	attachments.push_back(&a);
}

void Track::remove_attachment(TrackAttachment &a)
{
	AttachmentList::iterator i = find(attachments.begin(), attachments.end(), &a);
	if(i==attachments.end())
		throw key_error(&a);
	attachments.erase(i);
}

Track::AttachmentList Track::get_attachments_ordered(unsigned epi) const
{
	AttachmentList result = attachments;
	result.sort(AttachmentCompare(epi));
	return result;
}

void Track::save(list<DataFile::Statement> &st) const
{
	st.push_back((DataFile::Statement("position"), position.x, position.y, position.z));
	st.push_back((DataFile::Statement("rotation"), rotation.radians()));
	st.push_back((DataFile::Statement("tilt"), tilt.radians()));
	if(turnout_addr)
		st.push_back((DataFile::Statement("turnout_address"), turnout_addr));
	if(sensor_addr)
		st.push_back((DataFile::Statement("sensor_address"), sensor_addr));
	if(flex)
		st.push_back((DataFile::Statement("flex"), true));
}

void Track::save_dynamic(list<DataFile::Statement> &st) const
{
	if(turnout_addr)
		st.push_back((DataFile::Statement("path"), active_path));
}

void Track::turnout_event(unsigned id, unsigned state)
{
	if(id==turnout_id)
	{
		active_path = state;
		path_changing = false;
		signal_path_changed.emit(active_path);
	}
}

void Track::turnout_failed(unsigned id)
{
	if(id==turnout_id)
	{
		path_changing = false;
		layout.emergency(block, "Turnout failed");
	}
}


Track::Loader::Loader(Track &t):
	DataFile::ObjectLoader<Track>(t)
{
	add("path",       &Loader::path);
	add("position",   &Loader::position);
	add("rotation",   &Loader::rotation);
	add("tilt",       &Loader::tilt);
	add("turnout_id", &Loader::turnout_address);
	add("turnout_address", &Loader::turnout_address);
	add("sensor_id",  &Loader::sensor_address);
	add("sensor_address",  &Loader::sensor_address);
	add("flex",       &Track::flex);

	// deprecated
	add("slope",      &Loader::slope);
}

void Track::Loader::path(unsigned p)
{
	obj.set_active_path(p);
	if(obj.path_changing)
	{
		obj.active_path = p;
		obj.signal_path_changed.emit(p);
	}
}

void Track::Loader::position(float x, float y, float z)
{
	obj.set_position(Vector(x, y, z));
}

void Track::Loader::rotation(float r)
{
	obj.set_rotation(Angle::from_radians(r));
}

void Track::Loader::sensor_address(unsigned addr)
{
	obj.set_sensor_address(addr);
}

void Track::Loader::slope(float s)
{
	obj.set_tilt(Geometry::atan(s/obj.type.get_path_length(0)));
}

void Track::Loader::tilt(float t)
{
	obj.set_tilt(Angle::from_radians(t));
}

void Track::Loader::turnout_address(unsigned addr)
{
	obj.set_turnout_address(addr);
}

} // namespace R2C2
